#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import functools

def extend_to_collections(simple_func):
    @functools.wraps(simple_func)
    def collection_func(arg, *args, **kwargs):
        if isinstance(arg, list):
            return [simple_func(a, *args, **kwargs) for a in arg]
        elif isinstance(arg, tuple):
            return tuple(simple_func(a, *args, **kwargs) for a in arg)
        elif isinstance(arg, set):
            return set(simple_func(a, *args, **kwargs) for a in arg)
        else:
            return simple_func(arg, *args, **kwargs)
    return collection_func


def extend_to_collections_dirs(simple_func):
    collection_func = extend_to_collections(simple_func)
    @functools.wraps(simple_func)
    def flexible_func(arg, *args, dict_keys=True, dict_vals=False, **kwargs):
        if isinstance(arg, dict):
            if dict_keys and not dict_vals:
                return {collection_func(k, *args, **kwargs): v for (k, v) in
                    arg.items()}
            elif dict_vals and not dict_keys:
                return {k: collection_func(v, *args, **kwargs) for (k, v) in
                    arg.items()}
            elif dict_vals and dict_keys:
                return {collection_func(k, *args, **kwargs): collection_func(v,
                        *args, **kwargs) for (k, v) in arg.items()}
            else:
                return arg
        return collection_func(arg, *args, **kwargs)
    return flexible_func


def ignore_first_arg(decorator):
    def new_decorator(func):
        @functools.wraps(func)
        def func2(firstarg, *args, **kwargs):
            pfunc = functools.partial(func, firstarg)
            return decorator(pfunc)(*args, **kwargs)
        return func2
    return new_decorator


#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import json, os
from abc import ABC, abstractmethod
from Dhelpers.launcher import launch
path = os.path

parentfolder = path.split(__file__)[0]
saved_folder = path.join(parentfolder,"layouts_imported_from_xkb")
try:
    os.makedirs(saved_folder, exist_ok=True)
except PermissionError:
    pass
    


def get_files(folder):
    for f in os.listdir(folder):
        fp = path.join(folder, f)
        if path.isfile(fp): yield f, fp


class LayoutBase(ABC):
    
    json_savefolder = saved_folder
    folders=[]
    file_ending = ""
    available_layouts = {}
    
    def __init__(self, name, folder=None, raise_error=True):
        self.name = name
        self.custom_folder = folder
        self.key_dict = {}
        self.key_dict = self.get_keydict()
        if not self.key_dict and raise_error: raise ValueError
    
    def sourcepath(self):
        full_name = self.name + self.file_ending
        l = (self.custom_folder,) if self.custom_folder else self.folders
        for folder in l:
            fp = path.join(folder, full_name)
            if path.isfile(fp): return fp
        raise FileNotFoundError
    
    @abstractmethod
    def get_keydict(self):
        raise NotImplementedError
        
    
    def save_json_dict(self, name=None, folder=None):
        if not folder: folder = self.json_savefolder
        if not name: name = self.name
        with open(path.join(folder, name + ".json"), "w") as f:
            json.dump(self.key_dict, f, indent=4)
    
    @classmethod
    def find_available(cls, folder=None):
        l = (folder,) if folder else cls.folders
        saved=[]
        available={}
        for folder in l:
            for name, fp in get_files(folder):
                name = path.splitext(name)[0]
                if name in saved: continue
                try:
                    obj = cls(name, folder=folder)
                except:
                    # if this object cannot be created, do not list it as
                    # available
                    continue
                saved.append(name)
                available[name] = obj
        cls.available_layouts = available
        return saved
    
    @classmethod
    def create_available_jsons(cls):
        for name,obj in cls.available_layouts.items():
            obj.save_json_dict()
    
    
    def __lt__(self, other):
        if not isinstance(other, self.__class__): return NotImplemented
        d = {}
        kd2 = self.key_dict
        kd1 = other.key_dict
        for key in kd2:
            right = kd2[key][0]
            try:
                left = kd1[key][0]
            except KeyError:
                left = key
            if left is not right: d[left] = right
        return d
    
    




class Layout(LayoutBase):
    
    folders = [saved_folder]
    file_ending = ".json"
    
    def get_keydict(self):
        with open(self.sourcepath()) as f: return json.load(f)
    
    
    
    
class Layout_from_klfc(LayoutBase):
    
    klfc_option = "from-???"
    
    def get_keydict(self):
        json_str = launch.get(
                f"klfc --{self.klfc_option} '{self.sourcepath()}'")
        return self.convert_from_klfc_json(json.loads(json_str))
    
    @staticmethod
    def convert_from_klfc_json(klfc_dict):
        d = {}
        for entry in klfc_dict["keys"]:
            d[entry["pos"].lower()] = entry["letters"]
        return d
    
    
class XKBLayout(Layout_from_klfc):
    
    folders = ["/usr/share/X11/xkb/symbols"]
    klfc_option = "from-xkb"

    


#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from .launcher import *
from .arghandling import *
from .baseclasses import *
from .customprint import *
from .decorators import *
from .strhandling import *
from .adaptor import *
from .this_script import *
from .counting import *
from .linux_checks import *
from .external import *
from .KeyboardLayouts import *
from .named import *
from . import *
from .container import *


#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import time, threading, importlib
from .container import container

def count(key, dic, waittime, maxc=None):
    """
    This function simply counts how often it is called within waittime and
    returns that number.
    """
    key = str(key)
    c = dic.get(key, 0)  # dic must be a globally defined dictionary
    if maxc and (c >= maxc):
        dic[key] = 0
        return 0
    dic[key] = c + 1
    threading.Thread(target=uncount,args=(key, dic, waittime)).start()
    return c

def uncount(key, dic, waittime):
    time.sleep(waittime)
    if dic[key] > 0:
        dic[key] -= 1

container.dpress_store = {}

def dpress(keyname, timebetween):
    return count(keyname, container.dpress_store, timebetween, maxc=2) == 1



#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from warnings import warn
import time, threading, subprocess, psutil, multiprocessing
from .arghandling import PositiveInt


def select_running_processes(*ps):
    # print(ps,type(ps))
    for p in ps:
        if isinstance(p, psutil.Process):
            if p.is_running(): yield p
        else:
            try:
                pid = p.pid
            except AttributeError:
                # print(p,type(p))
                if isinstance(p, PositiveInt):
                    pid = p
                else:
                    raise TypeError
            try:
                proc = psutil.Process(pid)
            except psutil.NoSuchProcess:
                pass
            else:
                if proc.is_running(): yield proc
   
def get_running_process(p):
    t = tuple(select_running_processes(p))
    if not t:
        raise ProcessLookupError("No running process for " + str(p))
    if len(t) > 1:
        raise ProcessLookupError("More than 1 running process found.")
    return t[0]



def terminate_process(*pids_or_processes, timeout=5, kill=True):
    procs = tuple(select_running_processes(*pids_or_processes))
    for proc in procs: proc.terminate()
    gone, alive = psutil.wait_procs(procs, timeout)
    if alive and kill:
        kill_process(*alive, timeout=timeout)
        
        
def kill_process(*pids_or_processes, timeout=5):
    procs = tuple(select_running_processes(*pids_or_processes))
    for proc in procs: proc.kill()
    gone, alive = psutil.wait_procs(procs, timeout)
    if alive:
        raise TimeoutError(f"The following process(es) could not be "
                           f"killed after {timeout} seconds:\n{alive}")


def find_other_instances(compare=("name", "exe", "cmdline"), ad_value=None):
    this_process = psutil.Process()
    this_dict = this_process.as_dict(compare, ad_value)
    print(this_dict)
    print("find other instances")
    for p in psutil.process_iter(compare, ad_value):
        if p == this_process: continue
        if this_dict["name"] == p.info["name"]:
            print(p.info)


class LaunchFuncs:
    """This class offers functions to introduce an easier syntax for calling
    subprocesses and threads. It should be used via the instance "launch".
    examples:
    launch("notify-send text")
    launch(["notify-send","text"])
    launch("notify-send","text")
    """
    
    @staticmethod
    def thread(targetfunc, *argstopass, initial_time_delay=None,
            **kwargstopass):
        if initial_time_delay:
            subthread = threading.Timer(initial_time_delay, targetfunc,
                    args=argstopass, kwargs=kwargstopass)
        else:
            subthread = threading.Thread(target=targetfunc,
                args=argstopass, kwargs=kwargstopass)
        subthread.start()
        return subthread
    
    
    
    PIPE = subprocess.PIPE
    CalledProcessError = subprocess.CalledProcessError
    TimeoutExpired = subprocess.TimeoutExpired
    SubprocessError = subprocess.SubprocessError
    class SubprocessStdErrReceived(SubprocessError):
        pass
    class Automatic:
        pass
    
    @staticmethod
    def _handle_arguments(*args):
        l = len(args)
        if l == 0:
            raise SyntaxError("no arguments for Popen specified.")
        elif l == 1:
            args = args[0]
            if isinstance(args, str):
                if " " in args: return args, True
                return args, False
            elif not isinstance(args, (list, tuple)):
                raise TypeError(
                        "command passed must be a string, list or tuple.")
        return [str(i) for i in args], False
    
    default_Popen_kwargs = {"universal_newlines": True}
    
    def set_default_Popen_kwargs(self, **kwargs):
        self.default_Popen_kwargs = kwargs
    
    def __call__(self, *args, stdout=None, stderr=None, stdin=None, check=False,
            check_err=False, timeout=None, shell=Automatic, wait=False,
            couple=False, **Popen_kwargs):
        """Does NOT wait for the process to finish before continuing."""
        args, s = self._handle_arguments(*args)
        if shell is not self.Automatic: s = shell
        if stdout is True: stdout = subprocess.PIPE
        if stderr is True: stderr = subprocess.PIPE
        if stdin is True: stdin = subprocess.PIPE
        if stderr is None and check_err: stderr = subprocess.PIPE
        kwargs = self.default_Popen_kwargs.copy()
        if Popen_kwargs: kwargs.update(Popen_kwargs)
        Popen_object = psutil.Popen(args, shell=s, stdout=stdout, stderr=stderr,
                stdin=stdin, **kwargs)
        Popen_object.time_started = time.time()
        if couple:
            Popen_object.couple_object = self.CoupleProcess(Popen_object)
        if wait:
            self.wait_check_process(Popen_object, check, check_err, timeout,
                    raise_error=True)
        elif check or timeout or check_err:
            self.thread(self.wait_check_process, Popen_object, check, check_err,
                    timeout, raise_error=False)
        return Popen_object
    
    @staticmethod
    def wait_check_process(Popen_object, check_exitCode=True, check_err=True,
            timeout=None, raise_error=False):
        try:
            exitCode = Popen_object.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            Popen_object.kill()
            Popen_object.wait()
            raise
        try:
            Popen_object.duration = time.time() - Popen_object.time_started
        except AttributeError:
            pass
        if check_exitCode and exitCode:
            msg = "Command '{Popen_object.args}' returned Nonzero exit status " \
                  "" \
                  "{exitCode} after duration {" \
                  "Popen_object.duration}.".format_map(locals())
            if raise_error: raise subprocess.SubprocessError(msg)
            warn(msg)
        if check_err:
            err = Popen_object.stderr.read()
            if err:
                msg = "Command '{Popen_object.args}' returned following value " \
                      "" \
                      "for stderr after duration {Popen_object.duration}:\n{" \
                      "err}".format_map(locals())
                if raise_error: raise subprocess.SubprocessError(msg)
                warn(msg)
    
    
    
    def get(self, *args, stdout=True, stderr=True, stdin=None, ret_stdout=True,
            check=True, check_err=True, shell=Automatic, input=None,
            **Popen_run_kwargs):
        """Waits for the subprocess to finish and returns the
        subproces.CompletedProcess object."""
        args, shell_ = self._handle_arguments(*args)
        if shell is not self.Automatic: shell_ = shell
        if stdout is True: stdout = subprocess.PIPE
        if stderr is True: stderr = subprocess.PIPE
        if stdin is True: stdin = subprocess.PIPE
        if stdout is not subprocess.PIPE: ret_stdout = False
        if stderr is not subprocess.PIPE: check_err = False
        if not check_err and ret_stdout:
            # in this case, the stderr value is not used anyway
            stderr = None
        time_started = time.time()
        # _print(args)
        if input is not None:
            Popen_run_kwargs["input"] = input
        elif stdin is not None:
            Popen_run_kwargs["stdin"] = stdin
        kwargs = self.default_Popen_kwargs.copy()
        kwargs.update(Popen_run_kwargs)
        CompletedProc = subprocess.run(args, shell=shell_, stdout=stdout,
                stderr=stderr, check=check, **kwargs)
        CompletedProc.duration = time.time() - time_started
        if check_err and CompletedProc.stderr:
            raise self.SubprocessStdErrReceived(
                    "Command '{args}' returned following value for stderr "
                    "after duration {CompletedProc.duration}:"
                    "{CompletedProc.stderr}".format_map(locals()))
        if ret_stdout:
            stdout = CompletedProc.stdout
            if stdout.endswith("\n"): stdout = stdout[:-1]
            return stdout
        return CompletedProc
    
    
    def wait(self, *args, stdout=None, stderr=None, check_err=False, **kwargs):
        if stderr is None:
            if check_err: stderr = subprocess.PIPE
        else:
            check_err = False
        return self.get(*args, stdout=stdout, stderr=stderr,
                check_err=check_err, ret_stdout=False, **kwargs)
    
    
    class CoupleProcess:
        def __init__(self, child, parent=None, checktime=1):
            self.checktime = checktime
            self.child = get_running_process(child)
            if parent:
                self.parent = get_running_process(parent)
            else:
                # use this process as parent
                self.parent = psutil.Process()
            # print(self.parent,self.child)
            checker_process = multiprocessing.Process(target=self.run)
            checker_process.daemon = True
            checker_process.start()
        def run(self):
            time.sleep(self.checktime)
            while self.parent.is_running() and self.child.is_running() and \
                self.parent.status() != psutil.STATUS_ZOMBIE  and  \
                    self.child.status() != psutil.STATUS_ZOMBIE:
                time.sleep(self.checktime)
            #print(self.parent.status(), self.child.status())
            kill_process(self.child, self.parent)



launch = LaunchFuncs()


#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import os, psutil, __main__, time
from .external import ExternalDict
from .launcher import terminate_process, launch

path = os.path


def check_process_exists(pid, ctime):
    try:
        process = psutil.Process(pid)
    except psutil.NoSuchProcess:
        pass
    else:
        if process.create_time() == ctime:
            return process
    return False



class ProcessCheckerExternalDict(ExternalDict):
    
        
    def load_profile(self, profile=None):
        d = super().load_profile(profile)
        for pid, ctime in d.copy().items():
            p = check_process_exists(pid, ctime)
            if p is False:
                d.pop(pid)
                self.empty_profile(profile,pid)
            else:
                self.processes.append(p)
        return d
    
    def get_val(self):
        self.processes = []
        super().get_val()
        return self.processes



class ThisScript:
    
    def __init__(self, name=None, folder = None, register=True, log_folder=None):
        if not name or not log_folder:
            try:
                f = __main__.__file__
            except AttributeError:
                f = None
        if name:
            self.name = name
        else:
            if not f:
                raise NameError("Name argument of ThisScript not specified and "
                                "no __main__.__file__ attribute found.")
            self.name = path.splitext(path.basename(f))[0]
        
        self.external_dict = ProcessCheckerExternalDict("Script_processes",
                folder=folder, profile=self.name, incl_default_profiles=False)
        self.process = psutil.Process()
        if register:
            self.register()
            launch.thread(self.register_thread)

    @property
    def active_processes(self):
        return self.external_dict.load()

    @property
    def registered_processes(self):
        return self.external_dict.val

    def register(self, pid=None):
        process = self.process if pid is None else psutil.Process(pid)
        self.external_dict.update_items({process.pid: process.create_time()})
        # identifying a process needs pid AND  # creation time. this is
        # actually also stored in  # process._ident and used for comparison
        # checks in the __eq__  # method of the process class of psutil
        self.external_dict.load()
    
    def register_thread(self):
        while True:
            time.sleep(10)
            self.register()
            
            

    def other_instance_procs(self):
        t=  tuple(proc
            for proc in self.active_processes if proc != self.process)
        if t: print(f"Found earlier instances: {t}")
        return t
    
    
    def terminate_other_instances(self, timeout=5):
        terminate_process(*self.other_instance_procs(), timeout=timeout)
        aprocs = self.active_processes
        assert len(aprocs) == 1
        assert aprocs[0] == self.process

    def terminate(self):
        terminate_process(self.process)


#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from Dhelpers.launcher import launch
from .baseclasses import TimedObject
import time, collections

class Container:
    
    def __init__(self):
        self.store = {}
        self.active_object_lists = collections.defaultdict(list)
    
    def set_temp_store_key(self,keyname, value, timeout=60*30):
        self.store[keyname] = value
        if timeout:
            launch.thread(self.remove_temp_store_key, keyname, timeout)
    
    def remove_temp_store_key(self,keyname, timeout=0):
        time.sleep(timeout)
        if keyname in self.store: self.store.pop(keyname, None)
    
    
    def key_stored(self, *args, **kwargs):
        return key_stored(self, *args,**kwargs)
    
    
class key_stored(TimedObject):
    def __init__(self, container_instance, keyname, timeout=60*30, value=None):
        super().__init__(timeout=timeout)
        self.keyname = keyname
        self.value = value
        self.value_before = None
        self.container = container_instance
        self.store = container_instance.store
        self.active_objects = container_instance.active_object_lists[keyname]
    
    def _start_action(self):
        self.value_before = self.store.get(self.keyname)
        self.active_objects.append(self)
        if self.value is not None:
            self.store[self.keyname] = self.value
        return self.active_objects
    
    def _stop_action(self):
        ao = self.active_objects
        if self not in ao: raise ValueError
        last = ao[-1]
        if last is self:
            if self.value_before is not None:
                self.store[self.keyname] = self.value_before
            else:
                try:
                    self.store.pop(self.keyname)
                except KeyError:
                    pass
        else:
            for i in range(len(ao)):
                if ao[i] is self: break
            ao[i+1].value_before = self.value_before
        ao.remove(self)
    
    def __repr__(self) -> str:
        sup = super().__repr__()[:-1]
        return sup + f" (key:{self.keyname}, value:{self.value}, " \
                     f"value_before:{self.value_before}) >"



container = Container()


#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#


import importlib, inspect, functools, types, os, pkgutil, sys, logging, \
    traceback, warnings, operator
from .baseclasses import RememberInstanceCreationInfo, KeepInstanceRefs
from .arghandling import (check_type, remove_first_arg, ArgSaver)



class AdaptionError(Exception):
    pass


def adaptionmethod(target_name=None, require=False):
    """this is a decorator to turn methods into adaptionmethods"""
    if callable(target_name):
        # if target_name is callable that means that @adaptionmethod was used
        #  without argument. in this case, target_name is actually the
        # decorated method, and we directly return the result of decoration
        return AdaptionFuncPlaceholder(cls_func=target_name, target_name=None,
                require_target=require)
    elif target_name is None or isinstance(target_name, str):
        # in this case we want to return another decorator to apply to the
        # cls_func:
        return functools.partial(AdaptionFuncPlaceholder,
                target_name=target_name, require_target=require)
    raise TypeError



class AdaptionFuncPlaceholder:
    def __init__(self, cls_func, target_name, require_target):
        self.cls_func = cls_func
        self.__name__ = cls_func.__name__
        self.target_name = self.__name__ if target_name is None else target_name
        self.target_modifier_func = None
        self.require_target = require_target
        self.modifier_argnum = None
        self.__signature__ = inspect.signature(cls_func)
    
    def __call__(self, adaptioninstance, *args, **kwargs):
        """ Usage example: HookAdaptor.keyb(instance,*args,**kwargs) instead
        of instance.keyb(*args,**kwargs)"""
        check_type(AdaptorBase, adaptioninstance)
        amethod = getattr(adaptioninstance, self.__name__)
        check_type(AdaptionMethod, amethod)
        return amethod(*args, **kwargs)
    
    def target_modifier(self, func):
        """This is a decorator to specify a function as target_modifer."""
        if not callable(func): raise TypeError
        self.target_modifier_func = func
        self.modifier_argnum = len(inspect.signature(func).parameters)
        if not (2 <= self.modifier_argnum <= 3):
            raise SyntaxError("2 or 3 parameters required for target_modifier"
                              "func: self, target(, adaptionmethod)")
        if func.__defaults__ is not None:
            raise SyntaxError("""target_modifer func must not have any
            default argument values.""")
        return func



class AdaptionMethod:
    def __init__(self, placeholder, Adaptor_instance):
        check_type(AdaptionFuncPlaceholder, placeholder)
        check_type(AdaptorBase, Adaptor_instance)
        self.placeholder = placeholder
        self.cls_func = placeholder.cls_func
        functools.update_wrapper(self, self.cls_func)
        # this will copy __doc__, __name__ and __module__, as well as signature
        self.Adaptor_instance = Adaptor_instance
        self.__self__ = Adaptor_instance
        self.impl_info = None
        self.target_space = None
        self.dependency_dict = None
        self.target = None
        self.target_name = placeholder.target_name
        self.target_modifier = placeholder.target_modifier_func
        self.require_target = placeholder.require_target
        self.target_signature = None
        self.target_param_num = None
        
        self.__signature__ = remove_first_arg(self.cls_func)
        
        # print(self.__name__, self.signature, self.signature_to_pass)
        
        self._called_args = None
        self._called_kwargs = None
    
    
    
    def adapt(self, impl_info):
        impl = self.Adaptor_instance.implementation
        name = self.__name__
        impl.update_target_spaces(method_infos={name: impl_info})
        self.set_target()
        return self.target_space
    
    
    def __call__(self, *args, **kwargs):
        self._called_args = args
        self._called_kwargs = kwargs
        # logging.info("  Method %s of %s called from:\n%s%s", self.__name__,
        #       self.Adaptor_instance.__class__.__name__,
        #      *traceback.format_stack(limit=3)[:2])
        if not self.Adaptor_instance.is_adapted:
            # this will automatically adapt the Adaptor_instance if
            # adapt_on_first_use is True
            raise AdaptionError("No implementation chosen for "
                                "following adaptor:\n%s"%self.Adaptor_instance)
        try:
            ret = self.cls_func(self.Adaptor_instance, *args, **kwargs)
        except TypeError as e:
            if str(e).endswith("object is not callable") and self.target in (
                    None, NotImplemented):
                raise AdaptionError("target of adaptionmethod %s is "
                                    "%s."%(self.__name__, self.target)) from e
            else:
                raise
        self._called_args = None
        self._called_kwargs = None
        return ret
    
    # Usual execution frames for AdaptionMethods:
    #  __call__ ( saves arguments)
    # --> cls_func (=function decorated with @adaptionmethod)
    # --> target_with_args (can automatically pass arguments)
    # --> target
    
    def target_with_args(self):
        if self.target_param_num == 0: return self.target()
        if self._called_args is None:
            raise SyntaxError("""target_with_args() called from outside
                                adaptionmethod.""")
        bound = self.__signature__.bind(*self._called_args,
                **self._called_kwargs)
        bound.apply_defaults()
        return self.target(*bound.args, **bound.kwargs)
    
    
    def set_target(self):
        impl = self.Adaptor_instance.implementation
        name = self.__name__
        target_space = impl.method_target_spaces.get(name)
        if target_space is None:
            self.target = None
            self.target_space = None
            self.impl_info = None
            return
        try:
            target = getattr(target_space, self.target_name)
        except AttributeError:
            if self.require_target:
                raise AdaptionError(f"Required function '{self.__name__}' "
                f"was not found.\ntartget namespace: "
                f" {target_space}\nimplementation: {impl}")
            target = NotImplemented
        self.target_space = target_space
        self.impl_info = impl.method_infos.get(name, impl.main_info)
        obj, ty = target, type(target)
        if ty is types.MethodType:
            def turn_to_func(method):
                return lambda *args, **kwargs: method(*args, **kwargs)
            obj = turn_to_func(target)
        if self.target_modifier:
            add = (self,) if self.placeholder.modifier_argnum == 3 else ()
            obj = self.target_modifier(self.Adaptor_instance, obj, *add)
        if obj is not target and callable(target):
            assert callable(obj)
            functools.update_wrapper(obj, target)
        if callable(obj):
            self.target_signature = inspect.signature(obj, follow_wrapped=False)
            self.target_param_num = len(self.target_signature.parameters)
            # print( self.target, self.target_signature, self.signature_to_pass)
        self.target = obj
        return self.target_space
    
    def __repr__(self):
        if self.target is None or self.target is NotImplemented:
            target = str(self.target)
        else:
            try:
                name = self.target.__name__
            except AttributeError:
                name = self.target_name
            try:
                mod = self.target.__module__
            except AttributeError:
                mod = str(self.target_space)
            target = mod + "." + name
        return "%s with target %s>"%(super().__repr__()[:-1], target)


# class AdaptionMethodContainer(AdditionContainer, basic_class=AdaptionMethod):
#     def __call__(self, *args, **kwargs):
#         return functools.reduce(operator.add, tuple(m(*args,**kwargs) for m
#             in self.members))


def _get_AdaptionFuncPlaceholders(cls):
    for name, obj in vars(cls).items():
        if isinstance(obj, AdaptionFuncPlaceholder):
            yield name
    for basecls in cls.__bases__:
        for name in _get_AdaptionFuncPlaceholders(basecls):
            try:
                obj = getattr(cls, name)
            except AttributeError:
                continue
            if isinstance(obj, AdaptionFuncPlaceholder):
                yield name
                

class AdaptorBase(KeepInstanceRefs):
    implementation_source = None  # set in first level subclass
    # adapt_on_first_use = False
    autoadapt_active = False
    dependency_folder = None  # set in first level subclass
    _subclass_level = 0
    adaptionmethod_names = set()
    
    
    def __init_subclass__(cls):
        cls._subclass_level += 1
        l = tuple(_get_AdaptionFuncPlaceholders(cls))
        if l:
            # this means that this subclass can actually create instances
            if cls._subclass_level < 2:
                raise SyntaxError("First-level Subclass of AdaptorBase is  "
                                  "not allowed to have adaptionmethods.")
            cls.adaptionmethod_names = set(l)
                #this will inherit names already defined
            cls.implementation_classes = None
            cls.added_impl_names = {}
            
    
    def __init__(self, main_info=None, *, group=None, _primary=False,
            **method_infos):
        
        super().__init__()
        
        check_type((str, int), group, allowed=(None,))
        self.primary = False
        self.instance_group = group
        if _primary is True:
            if group is None: self.instance_group = "default"
            prim_inst = self.get_primary_instance()
            if prim_inst:
                raise ValueError("Primary instance for this instance group "
                                 "already set:\n" + str(prim_inst))
            # the following is a way of conditionally subclassing:
            RememberInstanceCreationInfo.__init__(self)
            self.primary = True  # self.adapt_on_first_use = True #overrides
            # class default
        self.implementation = None
        
        for name in self.adaptionmethod_names:
            placeholder = getattr(self.__class__, name)
            amethod = AdaptionMethod(placeholder, self)
            setattr(self, name, amethod)
        
        if self.instance_group is None:
            if main_info or method_infos:
                self.adapt(main_info, **method_infos)
            else:
                self.instance_group = "default"
        
        if self.instance_group is not None:
            if main_info or method_infos: raise ValueError
            if self.autoadapt_active: self.adapt()
    
    
    
    def adaptionmethods(self):
        return tuple(getattr(self, name) for name in self.adaptionmethod_names)
    
    
    
    def adapt(self, main_info=None, *, warn=True, raise_error=True,
            check_adapted=True, **method_infos):
        """main_info can be string, Implementation or ArgSaver object."""
        try:
            if isinstance(main_info, Implementation):
                impl = main_info
                if method_infos: impl.update_target_spaces(None, method_infos)
            else:
                if main_info is None and not method_infos:
                    main_info = self.get_from_impl_dict()
                impl = Implementation(self.__class__, main_info, method_infos)
        except AdaptionError as e:
            if raise_error: raise
            if warn:
                text = f"Encountered an Exception trying to adapt\n{self}\n"
                for _ in range(10):
                    if str(e):
                        text += type(e).__name__ + ": " + str(e) + "\n"
                    e = e.__cause__
                    if e is None: break
                    text += "caused by\n"
                warnings.warn(text)
            return False
        if impl.main_info is None and not impl.method_infos:
            if check_adapted:
                raise AdaptionError("Given implementation information not "
                            "valid and none found in default implementations")
        self.implementation = impl
        for amethod in self.adaptionmethods(): amethod.set_target()
        return impl.show_target_spaces()
    
    
    
    def __repr__(self):
        if self.primary:
            r = RememberInstanceCreationInfo.__repr__(self)[
                :-1] + ", primary instance of group '"
        else:
            r = super().__repr__()[:-1] + ", instance group: '"
        r += str(self.instance_group) + "', implementation: " + str(
                self.implementation) + ">"
        return r
    
    
    @property
    def is_adapted(self):
        return self.implementation is not None
    
    @classmethod
    def deactivate_autoadapt(cls):
        cls.autoadapt_active = False
    
    @classmethod
    def activate_autoadapt(cls):
        cls.autoadapt_active = True
    
    
    @classmethod
    def _adapt_all(cls, omit_adapted_inst=False, raise_error=True, warn=True,
            impl_source=None):
        for inst in cls.get_instances():
            if inst.is_adapted and omit_adapted_inst: continue
            impl_info = inst.get_from_impl_dict(impl_source=impl_source)
            inst.adapt(impl_info, raise_error=raise_error, warn=warn,
                    check_adapted=False)
    
    @classmethod
    def adapt_all(cls, *args, **kwargs):
        cls._adapt_all(*args, **kwargs)
        for subcls in cls.__subclasses__(): subcls.adapt_all(*args, **kwargs)
    
    
    @classmethod
    def check_adapted(cls, raise_error=True, warn=True):
        text = "The following instance is not adapted: "
        missing_inst = []
        for subcls in [cls] + cls.__subclasses__():
            for inst in subcls.get_instances():
                if not inst.is_adapted:
                    if raise_error: raise ValueError(text + str(inst))
                    if warn: warnings.warn(text + str(inst))
                    missing_inst.append(inst)
        return missing_inst
    
    
    
    # @property
    # def implementation(self):
    #     if self._implementation is None and self.adapt_on_first_use and not \
    #             self.autoadapt_active:
    #         try:
    #             self.adapt()
    #         except AdaptionError:
    #             self.adapt_on_first_use = False
    #     return self._implementation
    
    @classmethod
    def update_group_implementation(cls, apply=True, **new_group_infos):
        if not hasattr(cls, "adaptionmethod_names"): raise SyntaxError
        for group, impl_info in new_group_infos.items():
            x = getattr(cls.implementation_source, cls.__name__)
            setattr(x, group, impl_info)
            if apply:
                for inst in cls._get_group_instances(group): inst.adapt()
    
    
    # @classmethod
    # def implementation_class(cls, impl_class):
    #     """a class decorator"""
    #     check_type(type, impl_class)
    #     cls.added_impl_names[impl_class.__name__] = impl_class
    #     if cls.implementation_classes is None: cls.implementation_classes =
    #     set()
    #     cls.implementation_classes |= {impl_class}
    #     return impl_class
    
    @classmethod
    def _get_group_instances(cls, group_or_instance="default"):
        group = group_or_instance.instance_group if isinstance(
                group_or_instance, cls) else group_or_instance
        check_type((str, int), group)
        for instance in cls.get_instances():
            if instance.instance_group == group: yield instance
    
    @classmethod
    def _get_primary_instance(cls, group_or_instance="default"):
        for instance in cls._get_group_instances(group_or_instance):
            if instance.primary is True: return instance
        return None
    
    @classmethod
    def _get_from_impl_dict(cls, group_or_instance="default",
            Adaptor_class_or_name=None, impl_source=None):
        if impl_source is None: impl_source = cls.implementation_source
        
        if Adaptor_class_or_name is None:
            Adaptor_subclass_name = cls.__name__
        elif issubclass(Adaptor_class_or_name, cls):
            Adaptor_subclass_name = Adaptor_class_or_name.__name__
        else:
            Adaptor_subclass_name = Adaptor_class_or_name
        check_type(str, Adaptor_subclass_name)
        
        try:
            subclass_info_class = getattr(impl_source, Adaptor_subclass_name)
        except AttributeError:
            return
        
        if group_or_instance is None:
            raise ValueError
        elif isinstance(group_or_instance, cls):
            group = group_or_instance.instance_group
        else:
            group = group_or_instance
        check_type(str, group)
        
        try:
            main_info = getattr(subclass_info_class, group)
        except AttributeError:
            return  # it's ok to return main_info as None
        check_type((str, tuple, list, Implementation, ArgSaver), main_info)
        return main_info
    
    def get_group_instances(self):
        return self._get_group_instances(self.instance_group)
    
    def get_primary_instance(self):
        return self._get_primary_instance(self.instance_group)
    
    def get_from_impl_dict(self, impl_source=None):
        return self._get_from_impl_dict(self.instance_group,
                impl_source=impl_source)
    
    # @classmethod
    # def import_dependency(cls, dependency_name):
    #     check_type(str, dependency_name)
    #     check_type(dict, cls.dependency_dict)
    #     # the following tries to find a user defined dependency module in the
    #     #  dependency dict:
    #     dependency_module = cls.dependency_dict.get(dependency_name)
    #     if not dependency_module:
    #         # otherweise look up the dependency in the default dependency
    #         folder
    #         try:
    #             dependency_package = cls.default_dependency_folder
    #         except AttributeError as e:
    #             raise AttributeError(
    #                     "Could not find default dependency package location. "
    #                     "This should be set by the {"
    #                     "cls}.default_dependency_package"
    #                     " class attribute.".format_map(locals())) from e
    #         dependency_module = importlib.import_module(
    #                 dependency_package + "." + dependency_name)
    #     return dependency_module
    
    @classmethod
    def print_all_infos(cls):
        for subclass in cls.__subclasses__():
            print("subclass:", subclass)
            path = os.path.dirname(inspect.getfile(subclass))
            for submodule in pkgutil.iter_modules([path]):
                name = submodule[1]
                if not name.startswith("adapt_"):
                    continue
                print("implementation:", name[6:])
                full_name = subclass.__module__ + "." + name
                # mod=sys.modules[subclass.__module__+"."+name]
                try:
                    mod = importlib.import_module(full_name)
                except NotImplementedError:
                    print("-- NotImplementedError")
                    continue
                try:
                    dfl = mod.dependency_full_name
                except AttributeError:
                    dfl = None
                else:
                    print("dependency module:", dfl)
            for i in subclass.get_instances():
                print("")
                print("instance: ", i)
                print("defined in:", i.creation_module)
                print("on line:", i.creation_line)
            print("")
            print("")
            
            
            
    @classmethod
    def coupled_class(cls):
        """Immediately creates a baseclass to be used"""
        return type(cls.__name__ + ".CoupledClass", (CoupledClass,),
                dict(adaptor_class = cls))



wrap = functools.wraps(AdaptorBase.adapt)
class CoupledClass:
    
    adaptor_class = None
    adaptor = None

    @classmethod
    @wrap
    def adapt(cls, *args, **kwargs):
        cls.adaptor = cls.adaptor_class(group=None)
        return cls.adaptor.adapt(*args, **kwargs)
        
    @wrap
    def adapt_instance(self, *args, **kwargs):
        self.adaptor = self.adaptor_class(group=None)
        return self.adaptor.adapt(*args, **kwargs)



class Implementation:
    
    def __repr__(self):
        s = f"<Implementation('{self.main_info}'"
        if self.method_infos:
            s += f", {self.method_infos})"
        s += f") for {self.adaptorcls.__name__}>"
        return s
    
    def __str__(self):
        s = str(self.main_info)
        if self.method_infos:
            s += f", {self.method_infos}"
        return s
    
    def __info__(self):
        return self.adaptorcls, self.main_target_space, \
            self.method_target_spaces
    
    def __eq__(self, other):
        if isinstance(other,self.__class__):
            return self.__info__() == other.__info__()
        return NotImplemented
    
    def __hash__(self):
        return hash(str(self.__info__()))

    def show_target_spaces(self):
        if self.method_infos: return self.method_target_spaces
        return self.main_target_space
    
    
    def __init__(self, adaptorclass=None, main_info=None, method_infos={}):
        if not issubclass(adaptorclass, AdaptorBase): raise TypeError
        self.adaptorcls = adaptorclass
        self.dependency_folder = adaptorclass.dependency_folder
        self.main_target_space = None
        self.method_target_spaces = {}
        self.main_info = None  # will be set in update method
        self.method_infos = {}  # will be set in update method
        self.update_target_spaces(main_info, method_infos)
    
    def update_target_spaces(self, main_info=None, method_infos={}):
        main_info, method_infos = self._resolve_arguments(main_info,
                method_infos)
        for name in method_infos:
            if name not in self.adaptorcls.adaptionmethod_names:
                raise NameError(f"Non-valid name for adaptionmethod of class "
                f"{self.adaptorcls}: {name}.")
            check_type((str, list, tuple), method_infos[name])
        if main_info is not None:
            check_type((str, list, tuple), main_info)
            self.main_target_space, self.main_info = self.get_targetspace(
                    main_info)
        for name in self.adaptorcls.adaptionmethod_names:
            if name in method_infos:
                m = method_infos[name]
                r1, r2 = self.get_targetspace(m)
                self.method_target_spaces[name] = r1
                self.method_infos[name] = r2
            elif self.main_target_space is not None:
                self.method_target_spaces[
                    name] = self.main_target_space
    # note: after using update_target_space (giving info arguments), the
    # self.method_target_spaces will contain all method names. while the
    # self.method_infos dict will only contain those infos explicitly  given.
    # This is why you should use  exactly the following expressions to query:
    # self_or_cls.method_target_spaces.get(name) and
    # self_or_cls.method_infos.get(name,self_or_cls.main_info)
    
    
    def refresh_target_spaces(self):
        return self.update_target_spaces(self.main_info, **self.method_infos)
    
    
    def _resolve_arguments(self, main_info, method_infos):
        old1, old2 = main_info, method_infos
        for _ in range(50):
            new1, new2 = self._find_arguments(old1, old2)
            if new1 == old1 and new2 == old2: break
            old1, old2 = new1, new2
        else:
            raise RecursionError
        return new1, new2
    
    def _find_arguments(self, main_info, method_infos):
        # returns the resolved main_info, method_infos pair
        if isinstance(main_info, self.__class__):
            new1, new2 = main_info.main_info, main_info.method_infos
        elif isinstance(main_info, ArgSaver):
            if len(main_info.args) == 1:
                new1 = main_info.args[0]
            elif len(main_info.args) == 0:
                new1 = None
            else:
                raise ValueError("Argsaver object has too many postional args.")
            new2 = main_info.kwargs
        else:
            return self._resolve_name(main_info), {k: self._resolve_name(v) for
                k, v in method_infos.items()}
        ret2 = new2.copy()
        ret2.update(method_infos)
        return new1, ret2
    
    def _resolve_name(self, obj):
        if isinstance(obj, (list, tuple)): return obj
        for _ in range(50):
            try:
                obj = self.adaptorcls.added_impl_names[obj]
            except KeyError:
                break
        else:
            raise RecursionError
        return obj
    
    def get_targetspace(self, info):
        if isinstance(info, str):
            # TODO: add option to give full path of module instead
            module_name = ".adapt_" + info
            module_location = self.adaptorcls.__module__
            module_full_name = module_location + module_name
            if self.dependency_folder:
                sys.path.insert(0, self.dependency_folder)
                # this makes sure that packages inside the dependency folder
                # are found first. Unless the module has already been imported.
            try:
                ret = importlib.import_module(module_full_name)
            except Exception as e:
                raise AdaptionError from e
            if self.dependency_folder: sys.path.pop(0)
            return ret, info
        elif isinstance(info, (list, tuple)):
            last_error = None
            for subinfo in info:
                try:
                    ret_tuple = self.get_targetspace(subinfo)
                except AdaptionError as e:
                    this_error = e.__cause__
                    if last_error: this_error.__cause__ = last_error
                    last_error = this_error
                    continue
                else:
                    return ret_tuple
            raise AdaptionError from last_error
        # this way, all the errors that lead here, will be visible.
        else:
            raise TypeError(
                    f"Require string, but got {info}.")  # TODO: Add option
            # for implementation classes


#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
def cut_right(string: str, keywords_to_cut, max_number_of_cuts=1):
    if type(keywords_to_cut) is str:
        keywords_to_cut = (keywords_to_cut,)
    for i in range(max_number_of_cuts):
        if string.endswith(keywords_to_cut):
            for keyword in keywords_to_cut:
                if string.endswith(keyword):
                    string = string[:len(keyword)]
                    break  # this breaks the inner for loop
        else:
            break  # stop checking if no keyword found anymore
    return string

def cut_left(string: str, keywords_to_cut, max_number_of_cuts=1):
    if type(keywords_to_cut) is str:
        keywords_to_cut = (keywords_to_cut,)
    for i in range(max_number_of_cuts):
        if string.startswith(keywords_to_cut):
            for keyword in keywords_to_cut:
                if string.startswith(keyword):
                    string = string[len(keyword):]
                    break  # this breaks the inner for loop
        else:
            break  # stop checking if no keyword found anymore
    return string



import re, unicodedata
from warnings import warn


# source: https://github.com/django/django/blob/master/django/utils/text.py

def slugify(value, allow_unicode=True):
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.
    """
    value = str(value)
    if allow_unicode:
        value2 = unicodedata.normalize('NFKC', value)
    else:
        value2 = unicodedata.normalize('NFKD', value).encode('ascii',
                'ignore').decode('ascii')
    value2 = re.sub(r'[^\w\s-]', '', value2.lower()).strip()
    ret = re.sub(r'[-\s]+', '-', value2)
    if ret =="":
        warn(f"Could not slugify following string: '{value}'")
        ret = value
    return ret



#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#

import collections, traceback, threading, time, weakref

from .arghandling import check_type
from Dhelpers.launcher import launch
from abc import ABC, abstractmethod
from inspect import signature

class KeepInstanceRefs:
    __instance_ref_dict__ = collections.defaultdict(list)
    # this class attribute is a dictionary, supposed to contain one list of
    # instance
    # references for each sub class.
    # {cls1: instance_ref_list1, cls2: instance_ref_list2, etc.}
    # if cls has no instances, simply return an empty list by default.
    
    def __init__(self):
        super().__init__() #remove?
        self.__instance_ref_dict__[self.__class__].append(weakref.ref(self))
        # self.__instance_ref_dict__[self.__class__] is the list containing
        # all instance
        # references for this given
        # sub-class
        # we need a weakref, so that we don't get errors if the instance
        # object is deleted
    
    @classmethod
    def get_instances(cls):
        instance_ref_list = cls.__instance_ref_dict__[cls]
        for inst_ref in instance_ref_list:
            inst = inst_ref()
            if inst is None:
                pass
                # print("deleting",inst)
                # instance_ref_list.remove(inst_ref) #remove dead weakrefs
                # automatically
            else:
                # print("found: ",inst)
                yield inst
          
    @classmethod
    def subclass_from_name(cls, name):
        subclasses = cls.__subclasses__()
        if name in subclasses: return name
        check_type(str, name)
        for subclass in subclasses:
            if subclass.__name__ == name: return subclass
    
    @classmethod
    def instance_num(cls):
        return sum(1 for _ in cls.get_instances())
        # this is equal to the generator's length
    
    
    
    
class UniqueInstances(KeepInstanceRefs):
    
    _error_msg = None
    
    def __init__(self):
        for inst in self.get_instances():
            if self == inst:
                raise MultipleInstanceError(self._error_msg, inst)
        KeepInstanceRefs.__init__(self)
        self._new_instance = True

    @classmethod
    def get_instance(cls, *args, **kwargs):
        try:
            self = cls(*args,**kwargs)
        except MultipleInstanceError as e:
            self = e.existing_inst
            self._new_instance = False
        if not hasattr(self,"_new_instance"): raise AttributeError
        return self
        
        
class MultipleInstanceError(Exception):
    def __init__(self, msg, existing_inst):
        super().__init__(msg)
        self.existing_inst = existing_inst
        
        
        
        
        

class InstanceCreationError(Exception):
    pass

class RememberInstanceCreationInfo:
    def __init__(self):
        for frame, line in traceback.walk_stack(None):
            varnames = frame.f_code.co_varnames
            if varnames is ():
                break
            if frame.f_locals[varnames[0]] not in (self, self.__class__):
                break
                # if the frame is inside a method of this instance,
                # the first argument usually contains either the instance or
                #  its class
                # we want to find the first frame, where this is not the case
        else:
            raise InstanceCreationError("No suitable outer frame found.")
        self.creation_module = frame.f_globals["__name__"]
        self.creation_file, self.creation_line, self.creation_function, \
            self.creation_text = \
            traceback.extract_stack(frame, 1)[0]
        self.creation_name = self.creation_text[
        :self.creation_text.find('=')].strip()
        threading.Thread(target=check_existence_after_creation,
                args=(self, frame)).start()
    
    def __repr__(self):
        try:
            r = super().__repr__()
        except TypeError:
            # this happens if not used as BaseClass, but called directly
            r = object.__repr__(self)
        return r[:-1] + " with creation_name '%s'>" %self.creation_name


# the following is better than a staticmethod, because it allows using the
# the __init__ method of RememberInstanceCreationInfo without subclassing it
def check_existence_after_creation(obj, frame):
    while frame.f_lineno == obj.creation_line:
        time.sleep(0.01)
    # this is executed as soon as the line number changes
    # now we can be sure the instance was actually created
    error = InstanceCreationError(
            "\nCreation name not found in creation frame.\n"
            "creation_file: %s \ncreation_line: %s \n"
            "creation_text: %s\ncreation_name (might be wrong): "
            "%s"%(obj.creation_file, obj.creation_line, obj.creation_text,
            obj.creation_name))
    nameparts = obj.creation_name.split(".")
    try:
        var = frame.f_locals[nameparts[0]]
    except KeyError as k:
        raise error from k
    try:
        for name in nameparts[1:]: var = getattr(var, name)
    except AttributeError as a:
        raise error from a
    if var is not obj: raise error



class TimedObject(ABC):
    
    def __init__(self, *args, timeout, duration=True, wait = False, **kwargs):
        self.timeout = timeout
        self._timeout_thread = None
        self.active = False
        self.start_flag = threading.Event()
        self.stop_flag = threading.Event()
        self.measure_duration = duration
        self.wait = wait
        self.exitcode = None
        if duration:
            self.starttime = None
            self.endtime = None
        super().__init__(*args,**kwargs)
    
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        cls.start.__signature__ = signature(cls._start_action)
        cls.stop.__signature__ = signature(cls._stop_action)
        

    def _timeout_action(self):
        # if this returns True, then the timeout will not be performed
        pass
    
    def _error_handler(self, exc_type, exc_val, exc_tb):
        # this can return True value to suppress exceptions inside the with
        # statement. by default they are re-raised
        pass
    
    
    @abstractmethod
    def _start_action(self,*args,**kwargs):
        raise NotImplementedError
    
    @abstractmethod
    def _stop_action(self, *args, **kwargs):
        raise NotImplementedError
    
    def start(self, *args, wait=None, **kwargs):
        if self.active:
            raise RuntimeError("TimedObject %s cannot be started as it "
                               "is already active." % str(self))
        self.active = True
        ret = self._start_action(*args,**kwargs)
        if ret is None: ret = self
        if self.measure_duration: self.starttime = time.time()
        if self.timeout:
            self._timeout_thread = launch.thread(self._timeout_stop,
                    initial_time_delay=self.timeout)
        self.start_flag.set()
        self.start_flag.clear()
        if wait is None and self.wait or wait is True: self.join()
        return ret
        
    def _timeout_stop(self):
        if self.active:
            if self._timeout_action() is True:
                self._timeout_thread = None
            else:
                self.stop("timeout")
                self._timeout_thread = True
    
    def stop(self, exitcode = None, *args, raise_errors = True, **kwargs):
        if not self.active and raise_errors:
            if self._timeout_thread is True:
                    raise TimeoutError("TimedObject %s can not be stopped as it has "
                        "already timed out after %s seconds." % (str(self),
                            str(self.timeout)))
            else:
                raise RuntimeError("TimedObject %s can not be stopped as "
                                   "it is not active."%str(self))
        try:
            self._timeout_thread.cancel()
        except:
            pass
        self._timeout_thread = None
        if self.active:
            self.active = False
            if self.measure_duration: self.endtime = time.time()
            self.exitcode = exitcode
            ret = self._stop_action(*args,**kwargs)
            self.stop_flag.set()
            self.stop_flag.clear()
            if ret is None and self.measure_duration: return self.duration()
            return ret


    def duration(self):
        if not self.measure_duration: raise ValueError
        if self.starttime is None: return 0
        t = self.endtime if self.endtime else time.time()
        return t-self.starttime



    def __enter__(self):
        return self.start(wait=False)

  
    def __exit__(self, *error_info):
        if self.active: self.stop("__exit__")
        return self._error_handler(*error_info)
    
    def join(self, timeout=None):
        if self.active is not True: return False
        self.stop_flag.wait(timeout)
        return True
        



class AdditionContainerAddend:

    ContainerClass = None

    def __add__(self, other):
        if isinstance(other,
                (self.ContainerClass.basic_class, self.ContainerClass)):
            return self.ContainerClass(self, other)
        try:
            return super().__add__(other)
        except AttributeError:
            return NotImplemented

    def __radd__(self, other):
        if other is 0: return self  # necessary for using sum()
        try:
            return super().__radd__(other)
        except AttributeError:
            return NotImplemented
        
    
    @property
    def members(self):
        return (self,)
       
    

class AdditionContainer:
    
    # must be set in the subclass:
    basic_class = None
    _methods_to_include = {}
    Addend = AdditionContainerAddend
    
    @property
    def members(self):
        return self._members
    
    def __init_subclass__(cls, basic_class=None, ordered=True):
        if basic_class is None:
            raise ValueError("Keyword argument basic_class required.")
        if not issubclass(basic_class, cls.Addend): raise TypeError
        cls.basic_class = basic_class
        cls.ordered=ordered
        basic_class.ContainerClass = cls
        for method_name, error_type in cls._methods_to_include.items():
            if not issubclass(error_type,Exception): raise TypeError
            setattr(cls, method_name, cls._create_combined_method(
                    method_name, error_type))
    
    
    def __init__(self, *args):
        self._members = []
        for arg in args:
            if isinstance(arg, self.__class__):
                self._members += arg._members
            elif isinstance(arg, self.basic_class):
                self._members += [arg]
            else:
                raise TypeError(arg)
    
    def __add__(self, other):
        if isinstance(other, (self.basic_class, self.__class__)):
            return self.__class__(self, other)
        try:
            return super().__add__(other)
        except AttributeError:
            return NotImplemented

    def __radd__(self, other):
        if other is 0: return self  # necessary for using sum()
        try:
            return super().__radd__(other)
        except AttributeError:
            return NotImplemented

    def __exit__(self, *error_info):
        booleans = tuple(bool(member.__exit__(*error_info)) for member in
                self._members)
        return True in booleans

    def __enter__(self):
        for m in self._members:
            m.__enter__()
        return self
    
    def __bool__(self):
        x = sum(bool(m) for m in self._members)
        return x > 0

        
    @staticmethod
    def _create_combined_method(method_name, error_type):
        def combined_method(self, *args, **kwargs):
            for member in self._members:
                try:
                    method = getattr(member,method_name)
                except AttributeError:
                    continue
                try:
                    return method(*args, **kwargs)
                except error_type:
                    continue
            raise error_type(f"Arguments {args, kwargs} not allowed for "
                    f"method {method_name} of AdditionObject {self}")
        return combined_method


#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from inspect import signature, Signature, Parameter as Param


class ArgSaver:
    def __init__(self,*args,**kwargs):
        self.args = args
        self.kwargs = kwargs



### changing signatures:

def _get_sig(sig_or_func):
    if isinstance(sig_or_func, Signature): return sig_or_func
    if callable(sig_or_func): return signature(sig_or_func)
    raise TypeError

def remove_first_arg(sig_or_func):
    sig = _get_sig(sig_or_func)
    args_without_first = list(sig.parameters.values())[1:]
    # #this removes the self argument from the cls_func parameters
    return sig.replace(parameters=args_without_first)

def add_kw_only_arg(sig_or_func, name,*,default=Param.empty, annotation=Param.empty):
    sig = _get_sig(sig_or_func)
    param = list(sig.parameters.values())
    new_param = Param(name, Param.KEYWORD_ONLY, default=default,
            annotation=annotation)
    if param:
        p = param[-1]
        if p.kind == Param.VAR_KEYWORD:
            param.insert(-1, new_param)
        else:
            param.append(new_param)
    else:
        param.append(new_param)
    return sig.replace(parameters=param)


# =======================================================================
# argument handling
# =======================================================================

def unpack_if_single(iterobj):
    if isinstance(iterobj, (list, tuple, set)):
        if len(iterobj) == 1:
            return iterobj[0]
    return iterobj

def pack_single(iterobj):
    if isinstance(iterobj, (list, tuple, set)):
        return iterobj
    return [iterobj]


def extract_if_single_collection(star_args):
    if len(star_args) == 1:
        x = star_args[0]
        if isinstance(x,(list, tuple, set, dict)):
            return x
    elif len(star_args) == 0:
        raise SyntaxError
    return star_args



# =======================================================================
# checking paramter types and values
# =======================================================================


def check_type(allowed_types, *objs, allowed=(), suppress_error=False):
    if len(objs) == 0: raise SyntaxError
    allowed_types = tuple(pack_single(allowed_types))
    # for a in allowed_types: if type(type(a)) is not type: # type(a) can be
    # a metaclass, but type(type(a)) is the equal to type
    # raise SyntaxError("allowed_types must contain a type or class.")
    for obj in objs:
        if isinstance(obj, allowed_types): continue
        if obj in allowed: continue
        break
    else:
        return True  # if all objs passed the test
    if suppress_error: return False
    raise TypeError("Got object %s of type "
                    "%s.\nallowed_types: %s\nallowed_additional:%s"%(
                        obj, type(obj), str(allowed_types), str(allowed)))


class _BoundedNumberMeta(type):
    def __instancecheck__(cls, instance):
        # print("checking",instance, cls.allowed_number_types)
        if isinstance(instance, cls.allowed_number_types):
            if cls.minval is not None:
                if instance < cls.minval: return False
            if cls.maxval is not None:
                if instance > cls.maxval: return False
            return True
        return False

# raise ValueError("Got obj %s of type %s.\nExpected minimum
# value: %s" % (
#                            str(obj), type(obj), min))
# raise ValueError("Got obj %s of type %s.\nExpected "
#                "maximum value: %s" % (
#                   str(obj), type(obj), max))

def BoundedNumber(minval=None, maxval=None, allowed_number_types=(int, float),
        name="BoundedNumberClass"):
    allowed_number_types = tuple(pack_single(allowed_number_types))
    return _BoundedNumberMeta(name, (), locals())

PositiveInt = BoundedNumber(minval=1, allowed_number_types=int,
        name="PositiveInt")
NonNegativeInt = BoundedNumber(minval=0, allowed_number_types=int,
        name="NonNegativeInt")
NonNegativeNumber = BoundedNumber(minval=0, name="NonNegativeNumber")



class _CollectionWithPropsMeta(type):
    def __instancecheck__(cls, instance):
        if not isinstance(instance, cls.allowed_collection_types): return False
        l = len(instance)
        if l < cls.minlen: return False
        if cls.maxlen is not None:
            if l > cls.maxlen: return False
        for item in instance:
            if cls.allowed_item_types is not None:
                if isinstance(item, cls.allowed_item_types): continue
            if item in cls.allowed_items: continue
            return False
        return True  # if all items passed the test

#
# raise ValueError("Got obj %s of type %s with length "
#                  "%s.\nExpected minimum length: %s" % (
#                      str(obj), type(obj), l, lmin))
# raise ValueError("Got obj %s of type %s with length "
#                                      "%s.\nExpected maximum length: %s" % (
#                                          str(obj), type(obj), l, lmax))

def CollectionWithProps(allowed_item_types=None, len=None, minlen=0,
        maxlen=None, allowed_items=(), allow_list=True, allow_tuple=True,
        allow_set=True, name="CollectionWithPropsClass"):
    if len: minlen, maxlen = len, len
    allowed_collection_types = \
        (list,)*allow_list + (tuple,)*allow_tuple + (set,)*allow_set
    return _CollectionWithPropsMeta(name, (), locals())



def make_tuple(obj, ty):
    if not isinstance(obj, (tuple, list, set)):
        if obj is None: return tuple()
        check_type(ty,obj)
        return (obj,)
    obj = tuple(obj)
    check_type(CollectionWithProps(ty),obj)
    return obj



#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import os,yaml, tempfile, __main__
path = os.path
import getpass,platform #useful for the profile functions
from Dhelpers.launcher import launch
from .strhandling import slugify

# taken from https://stackoverflow.com/questions/13293680/pyyaml-how-to-align-map-entries
# def dict_representer(dumper, dic):
#     keyWidth = max(len(k) for k in dic)
#     aligned = {k +' '*(keyWidth-len(k)):v for k,v in dic.items()}
#     return dumper.represent_mapping('tag:yaml.org,2002:map', aligned)

#yaml.add_representer(dict, dict_representer)

#==============================================================================
# #profile functions - return name of the profile of the given environment
def user_node():
    return getpass.getuser()+"_at_"+platform.node()

#==============================================================================

def yaml_write(obj, f):
    if isinstance(f, str):
        if not os.path.isfile(f):
            os.makedirs(os.path.dirname(f),exist_ok=True)
        with open(f,"w") as fp:
            yaml.dump(obj,fp, default_flow_style=False, indent=4)
    else:
        raise TypeError

def yaml_read(f):
    if isinstance(f, str):
        if path.isfile(f):
            with open(f) as fp: return yaml.load(fp)
        else:
            raise FileNotFoundError
    else:
        raise TypeError


def get_files(folder):
    try:
        dircontent = os.listdir(folder)
    except FileNotFoundError:
        return # creates an empty generator
    for f in dircontent:
        ff = path.join(folder, f)
        if path.isfile(ff): yield f, ff


class ExternalDict:
    basefolder = None
    _default_profile_dic = {"_": None}
    yaml_editor_command = None
        
    
    
    def __init__(self, name, folder=None, profile=None,
            incl_default_profiles=True):
        if folder is None: folder = self.basefolder #inherit from class
        if folder is None:
            try:
                f = __main__.__file__
            except AttributeError:
                raise Exception("No folder specified.")
            folder = path.join(path.dirname(path.realpath(f)), "external_dicts")
        if not os.access(folder, os.F_OK | os.W_OK | os.R_OK):
            raise Exception("No access to " + folder)
        if not os.path.isdir(folder):
            raise Exception(folder + " is not a folder")
        if isinstance(profile, str) or callable(profile):
            profile = [profile]
        elif profile is None:
            profile = []
        elif "global" in profile or "user_node" in profile:
            incl_default_profiles = False
        if incl_default_profiles:
            profiles = ["global"] + list(profile) + ["user_node"]
        else:
            profiles = list(profile)
        if not profiles: raise ValueError
        profile_values = []
        for p in profiles:
            if isinstance(p, str):
                if p=="user_node": p = user_node()
                profile_values.append(p)
            elif callable(p):
                profile_values.append(str(p()))
            else:
                raise ValueError(f"{p} is not a alid argument for profile.")
        self.profiles = profile_values
        self.basefolder = folder
        self.name = name
        self.folder = path.join(folder,name)
        os.makedirs(self.basefolder, exist_ok=True)
        for p in self.profiles:
            os.makedirs(self.path(p), exist_ok=True)
        self.load()
        self.locked = False
    
    
    def load(self):
        self.val = self.get_val()
        return self.val
    
    def get_val(self):
        val = dict()
        for profile in self.profiles:
            val.update(self.load_profile(profile))
        return val
        
        
    def load_profile(self,profile=None):
        d = dict()
        for fname, fpath in get_files(self.path(profile)):
            retrieved = yaml_read(fpath)
            if isinstance(retrieved, (tuple,list)):
                if retrieved[0] == "__key_val__":
                    if len(retrieved) != 3: raise ValueError
                    d[retrieved[1]] = retrieved[2]
                    continue
            d[fname] = retrieved
        return d
    
    def path(self, profile=None, *args):
        keys = tuple(slugify(arg) for arg in args)
        if profile is None: profile = self.profiles[-1]
        return path.join(self.folder, profile, *keys)
    
    def update_items(self,dic, profile=None, remove=False, keep_timestamps=True):
        if not dic: return
        if dic == self._default_profile_dic: return
        old_dic = self.load_profile(profile)
        #print(dic, "\n\n", old_dic)
        for key,val in dic.items():
            try:
                v_old = old_dic.pop(key)
            except KeyError:
                pass
            else:
                if val == v_old and keep_timestamps: continue
            #This makes sure that only files are touched if the value has
            #actually changed. This is cruical for synchronization by time stamp.
            yaml_write(["__key_val__",key,val], self.path(profile,key))
        if remove and old_dic: self.empty_profile(profile, *old_dic.keys())
        #print("removed", old_dic)
    
    def empty_profile(self, profile=None, *keys):
        if not keys:
            keys = tuple(fname for fname,fpath in get_files(self.path(profile)))
        for key in keys:
            #print(f"removing {key} from {profile}")
            os.remove(self.path(profile,key))
        
    def empty(self):
        for profile in self.profiles: self.empty_profile(profile)
    
    
    def edit(self, wait=False, keep_timestamps=True):
        if self.locked:raise RuntimeError(f"ExternalDict {self} already "
                    f"opened for editing")
        self.locked = True
        editable = dict()
        num = 0
        for profile in self.profiles:
            dic = self.load_profile(profile)
            if not dic: dic = self._default_profile_dic.copy()
            k=f"PROFILE {num} -- '{profile}'"
            editable[k] = dic
            num += 1
        editable[f"EXTERNAL OBJECT '{self.name}'"]={
            "basefolder":self.basefolder, }
        after_thread = launch.thread(self.after_edit, editable, keep_timestamps)
        if wait: after_thread.join()
        return after_thread
        
    def after_edit(self, editable, keep_timestamps):
        after_edit = self.edit_obj(editable)
        for k, dic in after_edit.items():
            if k.startswith("EXTERNAL OBJECT"): continue
            profile = k.split(" -- '")[-1][:-1]
            self.update_items(dic, profile, remove=True,
                    keep_timestamps=keep_timestamps)
        self.load()
        self.locked = False

    
    def edit_obj(self, obj):
        tempdir = tempfile.gettempdir()
        fpath = path.join(tempdir, f"external_edit_Dhelpers_{self.name}.yaml")
        if path.isfile(fpath): raise FileExistsError(fpath)
        yaml_write(obj, fpath)
        if not self.yaml_editor_command:
            raise ValueError("You need to set the "
                             "ExternalDict.yaml_editor_command attribute to "
                             "edit external objects with a text editor.")
        launch.wait(self.yaml_editor_command, fpath)
        new = yaml_read(fpath)
        os.remove(fpath)
        return new
        
    def __info__(self):
        return self.name, self.folder, self.profiles
    
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__info__() == other.__info__()
        return NotImplemented
    
    def __hash__(self):
        return hash(self.__info__())
    
    def __str__(self):
        return f"<{self.__class__.__name__}: {self.__info__()}>"


#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#


from setuptools import setup, find_packages
from time import sleep
import multiprocessing, os, subprocess, warnings, shutil, sys

dirname = os.path.dirname(os.path.realpath(__file__))
os.chdir(dirname)
module_folder = os.path.join(dirname,"Dlib")
sys.path.insert(0, module_folder)

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    subprocess.run(["licenseheaders", "-t", "licenseheader.txt", "-E", ".py"])
except FileNotFoundError as e:
    warnings.warn(str(e))




import Dpowers, Dhelpers, evdev_prepared

def my_setup(name, **kwargs):
    if "version" not in kwargs: raise ValueError
    shutil.rmtree(f"Dlib/{name}.egg-info", ignore_errors=True)
    shutil.rmtree("build", ignore_errors=True)
    print()
    print()
    print(f"------------------------  Package {name} ----------------------------")
    print()
    sleep(1)
    # for creating wheels using bdist_wheel it is somehow necessary to run
    # the setup command inside a seperate process.
    sub_process = multiprocessing.Process(target=setup_process, args=(name,),
            kwargs=kwargs)
    sub_process.start()
    sub_process.join()

def setup_process(name, **kwargs):
    setup(**kwargs,
        name=name,
        packages=find_packages("Dlib", include=[f"{name}*"]),
        package_dir={"": "Dlib"},
        include_package_data=False,
        ####must stay disabled to allow package_data option !!!
        python_requires=">=3.6", # metadata to display on PyPI
        author="DPS", author_email="dps@my.mail.de",
        url="https://github.com/dp0s/Dpowers",
        classifiers=["Programming Language :: Python :: 3",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"],
         )





my_setup("Dpowers",
    version = Dpowers.__version__,
    package_data = {"Dpowers.iconpower.icons" : ["*"]},

    install_requires=["pynput>=1.6.8", "pystray>=0.15",
        "Dhelpers==" + Dhelpers.__version__,
        "evdev_prepared>="+evdev_prepared.__version__],

    description="Unified Interface for automatic interaction",
    long_description=long_description,
    long_description_content_type="text/markdown"
        )



my_setup("Dhelpers",
    version = Dhelpers.__version__,
    package_data = {"Dhelpers.KeyboardLayouts.layouts_imported_from_xkb"
    : ["*"]},
    install_requires=["psutil>=5.4.2"],
    description="Dhelpers support modules"
    )

my_setup("evdev_prepared",
    version = evdev_prepared.__version__,
    install_requires=["evdev==1.2.0"],
    description="Useful classes to make usage of python evdev easier"
    )


print()
print("DONE")



#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#

from .baseclasses import KeepInstanceRefs
from .arghandling import check_type
from .decorators import (extend_to_collections)


class NamedObj(KeepInstanceRefs):
    defined_objects = None
    name_to_stnd_name = None
    names_with_important_capital_letters = None
    StandardizingDict = None
    defined_groups = None
    NameContainer = None
    
    __slots__ = ["names", "groups", "mapping_dict"]
    
    def __init_subclass__(cls, inherit_names=False):
        cls.names_with_important_capital_letters = set()
        cls.defined_groups = dict()
        cls.defined_objects = dict()
        cls.name_to_stnd_name = dict()
        sd = type(f"{cls.__name__}.StandardizingDict", (StandardizingDict,),{})
        sd.__module__ = cls.__module__
        sd.NamedClass = cls
        cls.StandardizingDict = sd
        new = NameContainer(cls)
        if inherit_names:
            for defining_class in cls.NameContainer.name_defining_classes:
                new.update(defining_class)
        cls.NameContainer = new
        
        
        
    @classmethod
    def update_names(cls, decorated_class):
        """A class decorator"""
        cls.NameContainer.update(decorated_class)
        return decorated_class
    
    def __init__(self, *names):
        super().__init__()
        self.groups = []
        self.names = list()
        self.add_names(*names)
        self.mappings = {}


    def add_to_group(self, group_name):
        if group_name in self.groups: return
        self.groups.append(group_name)
        try:
            members = self.defined_groups[group_name]
        except KeyError:
            self.defined_groups[group_name] = [self]
        else:
            if self not in members: members.append(self)
    
    def add_names(self, *names):
        #names = list(extract_if_single_collection(names))
        for name in names:
            if isinstance(name, int): name = str(name)
                #names[i] = name
                # names must be strings.
            check_type(str, name)
            if not self._check_if_name_is_allowed(name):
                raise NameError("Name " + name + " is forbidden.")
            try:
                existing_obj = self.instance(name)  # this uses make_comparable
            except KeyError:
                continue
            # if name was already used:
            if name in existing_obj.names:  # this does not use make_comparable
                raise NameError("Name " + name + " already assigned.")
            else:
                # this happens if e.g. agrave and Agrave are entered
                self.names_with_important_capital_letters.update({name.lower()})
        self.names += names
        self.calculate_refs()  # it is important that calculate_refs is run
            # each time so that names_with_important_capital_letters stays up
            # tp date
        
    @classmethod
    def calculate_refs(cls, ignore=None):
        cls.defined_objects = {}
        cls.name_to_stnd_name = {}
        for keyobj in cls.get_instances():
            if keyobj is ignore:
                continue  # this possibilitx is intended for the __del__ method
            for name in keyobj.names_comparable:
                if name in cls.defined_objects:
                    print(keyobj, keyobj.names,keyobj.names_comparable)
                    raise NameError("Key name " + str(
                            name) + " exists more than 1 time.")
                cls.defined_objects[name] = keyobj
                cls.name_to_stnd_name[name] = keyobj.name
            for group_name in keyobj.groups:
                members = cls.defined_groups[group_name]
                if keyobj not in members: members.append(keyobj)
        # for group in cls.defined_groups:
        #     group = cls.make_comparable(group)
        #     if group in cls.defined_objects:
        #         raise NameError(f"{group} defined multiple times for {cls}")
                
    
    @property
    def name(self):
        return self.names[0]
    
    @property
    def names_comparable(self):
        for name in self.names: yield self.make_comparable(name)
    
    @classmethod
    @extend_to_collections
    def make_comparable(cls, name):
        if isinstance(name, int): name = str(name)
        check_type(str, name)
        name = str(name)  #convert from Keyvent events to normal strings
        n = name.lower()
        if n not in cls.names_with_important_capital_letters:
            return n
            # usually return the lower case version.
            # but if the name exists in different cases,
            # return the name unchanged.
        return str(name)
    
    @property
    def stnd_name_comparable(self):
        return self.make_comparable(self.name)
    
    @classmethod
    def instance(cls, name):
        if not name: raise ValueError
        if isinstance(name, cls): return name
        return cls.defined_objects[cls.make_comparable(name)]

        
    
    @classmethod
    def get_stnd_name(cls, name):
        if isinstance(name, cls): return name.name
        return cls.name_to_stnd_name[cls.make_comparable(name)]
        
    # def set_stnd_name(self, stnd_name):
    #     self.stnd_name=stnd_name
    #     if self.stnd_name_comparable in self.stnd_names():
    #         raise Exception("This stnd name already given.")
    #     if self.stnd_name_comparable not in self.names_comparable:
    #         self.add_names(stnd_name)
    #     else:
    #         self.calculate_refs()
    
    
    @classmethod
    def _check_if_name_is_allowed(cls, name):
        # override this in a subclass
        return True
        # raise NotImplementedError("method '_check_if_name_is_allowed' must
        # be defined in a subclass of", cls)
    
    @classmethod
    def all_names_comparable(cls):
        return set(cls.name_to_stnd_name.keys())
    
    @classmethod
    def stnd_names(cls):
        return set(cls.name_to_stnd_name.values())
    
    def __del__(self):
        # 1. if an object is completely deleted and garbage collected,
        #   recalculate the refs
        # 2.  this will happen automatically after this instance is
        # completely destroyed
        # because the get_instances() method notices that the weakref is
        # targeting an non-existing object
        # 3. But for now, the instance is not yet destroyed,
        #  so we need to explicitely tell the calculate ref function to
        # ignore this instance
        #raise RuntimeError
        self.calculate_refs(ignore=self)
    
    
    def __eq__(self, other):
        if isinstance(other,int): other=f"[{other}]"
        if isinstance(other, str):
            return self.make_comparable(other) in self.names_comparable
        else:
            return super().__eq__(other)
    
    def __hash__(self):
        return hash(self.name)
    
    def __repr__(self):
        return super().__repr__()[:-1] + " with standard name '%s'>"%self.name
    
    def __str__(self):
        return f"{self.__class__.__name__}({self.name})"
    
    
    
class StandardizingDict:
    
    _running_number = 0
    NamedClass = None  #set by NamedObj init_subclass method
        
    def __init__(self, new_info=None, func=None, others_comparable=True):
        self.running_number = self._running_number
        self.__class__._running_number += 1
        self.other_dict = dict()
        self.registered_instances = dict()
        self.registered_names = dict()
        self.others_comparable= others_comparable
        self.make_comparable = self.NamedClass.make_comparable
        if new_info: self.update(new_info, func=func)
        
    
    def _comparable(self, k):
        return self.make_comparable(k) if self.others_comparable else k
        
            
    def copy(self):
        return self.__class__(self,others_comparable=self.others_comparable)
    
    def update(self, new_info, func=None):
        if isinstance(new_info, (set, list, tuple)):
            new_info = dict(zip(new_info, new_info))
        elif isinstance(new_info, StandardizingDict):
            new_info = new_info.normal_version()
        check_type(dict, new_info)
        if not func: func = lambda x: x
        for k,v in new_info.items(): self[k] = func(v)
    
    def normal_version(self, str_of_inst=False):
        if str_of_inst:
            f = lambda inst: str(inst)
        else:
            f = lambda inst: inst.name
        d = self.other_dict.copy()
        d.update({f(inst): inst.mappings[self.running_number]
            for inst in self.registered_instances})
        return d
    
    def __repr__(self) -> str:
        old = super().__repr__()[:-1]
        return old + f" with running number {self.running_number}>"
    
    def __eq__(self, other):
        if isinstance(other,self.__class__):
            if self.NamedClass == other.NamedClass and \
                self.other_dict == other.other_dict and \
                self.registered_instances == other.registered_instances:
                    return True
            return False
        return NotImplemented
    
    
    def __str__(self):
        return "StandardizingDict" + str(self.normal_version(
                str_of_inst=True)).replace("'","").replace('"','')
    
    
    def __setitem__(self, k, v) -> None:
        try:
            inst = self.NamedClass.instance(k)
        except KeyError:
            self.other_dict[self._comparable(k)] = v
        else:
            inst.mappings[self.running_number]=v
            for name in inst.names:
                self.registered_names[self.make_comparable(name)] = v
            self.registered_instances[inst] = v
    
    def __getitem__(self, k):
        # this method is optimized for speed
        try:
            if isinstance(k, self.NamedClass):
                return k.mappings[self.running_number]
            try:
                return self.registered_names[self.make_comparable(k)]
                #this is fast way to look up
            except KeyError:
                return self.other_dict[self._comparable(k)]
        except KeyError:
            raise KeyError(k)

        
    def __delitem__(self, k):
        try:
            inst = self.NamedClass.instance(k)
        except KeyError:
            del self.other_dict[self._comparable(k)]
        else:
            del inst.mappings[self._running_number]
            for name in inst.names: del self.registered_names[name]
            del self.registered_instances[inst]
        
    def get(self, k, default=None):
        try:
            return self[k]
        except KeyError:
            return default
        
    def pop(self, k, default=None):
        try:
            val = self[k]
        except KeyError:
            if default is not None: return default
            raise
        del self[k]
        return val
        
    def apply(self, k):
        try:
            inst = self.NamedClass.instance(k)
        except KeyError:
            try:
                return self.other_dict[self._comparable(k)]
            except KeyError:
                return k  #if k is not a defined name of any kind, return itself
        else:
            try:
                return inst.mappings[self.running_number]
            except KeyError:
                return inst.name
                #if k is not in dict, at least it can be standardized
    
    def __contains__(self, item):
        try:
            self[item]
            return True
        except KeyError:
            return False
    
    def keys(self):
        for inst in self.registered_instances: yield inst.name
        for key in self.other_dict.keys(): yield key
    
    def values(self):
        for value in self.registered_instances.values(): yield value
        for value in self.other_dict.values(): yield value
        
    def items(self):
        return zip(self.keys(), self.values())
    
    
    

def get_attributes(cls):
    for a, b in cls.__dict__.items():
        if not (a.startswith("__") and a.endswith("__")): yield (a,b)


class GroupContainer:
    def __call__(self, arg):
        return getattr(self,arg)
    
    def __init__(self, NamedClass):
        self.NamedClass =NamedClass
        
    def __getattr__(self, item):
        try:
            return self.NamedClass.defined_groups[item]
        except KeyError:
            raise AttributeError

class NameContainer:
    def __init__(self, NamedClass):
        self.NamedClass = NamedClass
        self.group = GroupContainer(NamedClass)
        self.name_defining_classes = []
        
    
    def update(self, NameDefiningClass):
        self.name_defining_classes.append(NameDefiningClass)
        self._iter_class(NameDefiningClass)
                
    def _iter_class(self,  cls, group_name=None, excluded=()):
        excluded = list(excluded) #creates copy
        for attr, val in get_attributes(cls):
            if val is None:
                excluded.append(attr)
                continue
            if attr in excluded: continue
            if isinstance(val,type):
                # nested class can be used to define a group
                gn = attr if group_name is None else group_name
                self._iter_class(val, group_name=gn, excluded=excluded)
            else:
                if isinstance(val, (tuple,list)):
                    names = list(val)
                elif isinstance(val,(str,int)):
                    names = [val]
                else:
                    raise TypeError(f"{val} of type {type(val)}")
                try:
                    inst = super().__getattribute__(attr)
                    # this will avoid the special __getattr__ from below
                except AttributeError:
                    inst = self.NamedClass(*names)
                setattr(self, attr, inst)
                if inst.names != names:
                    raise NameError(f"multiple defined attribute {attr}")
                if group_name: inst.add_to_group(group_name)
        for bcls in cls.__bases__:
            self._iter_class(bcls, group_name=group_name, excluded=excluded)
        
    
    def __getattr__(self, item):
        try:
            return self(item)
        except KeyError:
            raise AttributeError
        
    def __call__(self, name):
        return self.NamedClass.instance(name)
    
    
    
    
# class NameCombination:
#
#     def __init__(self, *named_subclasses):
#         for NamedClass in named_subclasses:
#             if not issubclass(NamedClass, NamedObj): raise TypeError
#         self.NamedClasses = named_subclasses
#         self.StandardizingDict = type(f"NameCombination.StandardizingDict",
#                 (StandardizingDict,),{})
#         self.StandardizingDict.NamedClass = self
#


#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
import builtins, inspect, traceback
from .container import container

# =======================================================================
# Alternating the builtin print function
# =======================================================================
# =======================================================================

# save the standard print function to _print:
if not inspect.isbuiltin(print):
    raise Exception("built in print function has already been changed.")
if hasattr(builtins, "_print"):
    if not builtins._print is builtins.print:
        raise Exception("_print already defined but is not the original print "
                        "function.")
else:
    builtins._print = builtins.print
    
container.execute_after_print = []

def find_original_print_func():
    if inspect.isbuiltin(builtins._print):
        return builtins._print
    elif inspect.isbuiltin(builtins.print):
        return builtins.print
    else:
        raise NameError("original print function not found.")

def restore_print_func():
    builtins.print = find_original_print_func()
    container.execute_after_print = []
    
    
class print_func_replaced:
    def __init__(self, temporary_print_func=None):
        self.temp_print = find_original_print_func() if temporary_print_func \
                                    is None else temporary_print_func
    def __enter__(self):
        self.saved = builtins.print, container.execute_after_print
        builtins.print = self.temp_print
        return self.saved[0]
    def __exit__(self, exc_type, exc_val, exc_tb):
        builtins.print, container.execute_after_print = self.saved

def execute_after_print(func):
    original_print = find_original_print_func()
    if not callable(func):
        raise TypeError
    if hasattr(container, "execute_after_print"):
        if func in container.execute_after_print:
            raise ValueError("function %s is already executed after print.")
        container.execute_after_print += [func]
    else:
        container.execute_after_print = [func]
    def _custom_print(*args, **kwargs):
        print_kwargs = {k: v for k, v in kwargs.items() if
            k in ("sep", "end", "file", "flush")}
        # this is necessary if additional kwargs are passed to custom_print
        original_print(*args, **print_kwargs)
        for fun in container.execute_after_print:
            fun(*args, **kwargs)
    builtins.print = _custom_print
    traceback.print = original_print  # this avoids problems with showing
    # errors


def print_last_traceback(*ignore, tb_lines=1, print_traceback=True, end="\n",
        **ignore_too):
    if not print_traceback:
        return
    original_print = find_original_print_func()
    frame = inspect.currentframe().f_back
    called_by_print = (frame.f_code.co_name == "_custom_print")
    if called_by_print:
        frame = frame.f_back
        tb_lines = container.default_tb_lines_to_print
    stack = traceback.extract_stack(frame, tb_lines)
    if called_by_print:
        file = stack[-1][0]
        if "IPython" in file and file.endswith(
                ("interactiveshell.py", "displayhook.py")):
            return
        if not end.endswith("\n"):
            original_print()  # create line between print output and traceback
    for line in traceback.format_list(stack):
        if line.endswith("\n"):
            line = line[:-1]
        original_print(line)
    original_print()  # to create a newline at the very end

def always_print_traceback(default_tb_lines=1):
    container.default_tb_lines_to_print = default_tb_lines
    execute_after_print(print_last_traceback)



#
#
# Copyright (c) 2020 DPS, dps@my.mail.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#
from Dhelpers.launcher import launch
from warnings import warn
import inspect

class BashDependencyError(Exception):
    pass

def check_bash_cmd(cmd, msg=None, expected_return=None, warning=False,
        run_command=None):
    module = inspect.currentframe().f_back.f_globals["__name__"]
    def warn_or_error(specific_msg):
        m = "\nTrying to load the following module: %s\n%s" % (module,specific_msg)
        if msg: m += "\n--> " + msg
        if warning:
            warn(m)
            return False
        else:
            raise BashDependencyError(m)

    first = cmd.split(" ")[0]
    try:
        out = launch.get("which "+ first)
    except launch.CalledProcessError:
        return warn_or_error(
                "The following required bash command could not be found: "
                "'%s'"%first)

    if (first != cmd and run_command is None) or run_command is True:
        try:
            out = launch.get(cmd)
        except launch.CalledProcessError:
            return warn_or_error("The test bash command '{cmd}' returned non "
                               "zero exit status.".format_map(locals()))
        
    if expected_return is not None and out != expected_return:
        return warn_or_error(
                "The test bash command '{cmd}'\n returned the value '{out}', "
                "\n but it should return '{expected_return}'.".format_map(
                        locals()))
    
    return out